# ==============================================================================
# Copyright 2022 Luca Della Libera.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Prior model."""

from typing import Any, Callable, Dict, Tuple, TypeVar

from torch import Tensor, nn
from torch.distributions import Distribution, Independent
from torch.nn import Module, Parameter

from bayestorch.models.utils import nested_stack


__all__ = [
    "PriorModel",
]


_T = TypeVar("_T", bound="PriorModel")


class PriorModel(Module):
    """Bayesian model that defines a prior over its parameters.

    Examples
    --------
    >>> import torch
    >>> from torch import nn
    >>>
    >>> from bayestorch.distributions import LogScaleNormal
    >>> from bayestorch.models import PriorModel
    >>>
    >>>
    >>> batch_size = 10
    >>> in_features = 4
    >>> out_features = 2
    >>> model = nn.Linear(in_features, out_features)
    >>> num_parameters = sum(parameter.numel() for parameter in model.parameters())
    >>> model = PriorModel(
    ...     model,
    ...     prior_builder=LogScaleNormal,
    ...     prior_kwargs={
    ...         "loc": torch.zeros(num_parameters),
    ...         "log_scale": torch.full((num_parameters,), -1.0),
    ...     },
    ... )
    >>> input = torch.rand(batch_size, in_features)
    >>> outputs, log_priors = model(input)

    """

    prior: "Distribution"
    """The prior distribution."""

    # override
    def __init__(
        self,
        model: "Module",
        prior_builder: "Callable[..., Distribution]",
        prior_kwargs: "Dict[str, Any]",
    ) -> "None":
        """Initialize the object.

        Parameters
        ----------
        model:
            The model.
        prior_builder:
            The prior builder, i.e. a callable that receives
            keyword arguments and returns a prior.
        prior_kwargs:
            The keyword arguments to pass to the prior builder.
            Tensor arguments are internally registered as
            parameters if their `requires_grad` attribute
            is True, as persistent buffers otherwise.

        Raises
        ------
        ValueError
            If an invalid argument value is given.

        """
        super().__init__()
        self.model = model
        self.prior_builder = prior_builder
        self.prior_kwargs = prior_kwargs = {
            k: v for k, v in prior_kwargs.items()
        }  # Avoid side effects

        # Build prior
        self.prior = self._build_distribution("prior", prior_builder, prior_kwargs)

    # override
    def to(self, *args: "Any", **kwargs: "Any") -> "_T":
        self.model.to(*args, **kwargs)
        super().to(*args, **kwargs)

        # Rebuild prior with parameters on new device
        self.prior = self._build_distribution(
            "prior", self.prior_builder, self.prior_kwargs
        )

        return self

    # override
    def forward(self, *args: "Any", **kwargs: "Any") -> "Tuple[Any, Tensor]":
        """Forward pass.

        In the following, let `B = {B_1, ..., B_k}` denote the
        batch shape and `O = {O_1, ..., O_m}` the shape of a
        leaf value of the underlying model output (can be a
        nested tensor).

        Parameters
        ----------
        args:
            The positional arguments to pass to the underlying model.
        kwargs:
            The keyword arguments to pass to the underlying model.

        Returns
        -------
            - The outputs, shape of a leaf value: ``[1, *B, *O]``;
            - the log priors, shape: ``[1]``.

        """
        # Extract particle
        particle = nn.utils.parameters_to_vector(self.model.parameters())

        # Compute log priors
        log_priors = self.prior.log_prob(particle)[None]

        # Forward pass
        outputs = [self.model(*args, **kwargs)]

        return nested_stack(outputs), log_priors

    def _build_distribution(
        self,
        name: "str",
        distribution_builder: "Callable[..., Distribution]",
        distribution_kwargs: "Dict[str, Any]",
    ) -> "Distribution":
        # Extract particle
        # particle = model parameters flattened into a 1D vector
        particle = nn.utils.parameters_to_vector(self.model.parameters())

        # Build distribution
        for k, v in distribution_kwargs.items():
            key = f"{name}_{k}"
            if isinstance(v, Tensor):
                if key in self._parameters:
                    v = self._parameters[key]
                elif key in self._buffers:
                    v = self._buffers[key]
                else:
                    v = self._register_tensor(key, v.to(particle))
            distribution_kwargs[k] = v
        distribution = distribution_builder(**distribution_kwargs)

        # Adjust distribution shape
        batch_ndims = len(distribution.batch_shape)
        if batch_ndims > 0:
            distribution = Independent(distribution, batch_ndims)

        # Validate distribution event shape
        event_shape = distribution.event_shape
        if event_shape != (1,) and event_shape != particle.shape:
            raise ValueError(
                f"{name.capitalize()} event size ({event_shape.numel()}) must be "
                f"equal to the number of model parameters ({particle.numel()})"
            )

        return distribution

    def _register_tensor(self, name: "str", input: "Tensor") -> "Tensor":
        if input.requires_grad:
            input = Parameter(input)
            self.register_parameter(name, input)
        else:
            self.register_buffer(name, input)
        return input

    # override
    def __repr__(self) -> "str":
        return f"{type(self).__name__}(model: {self.model}, prior: {self.prior})"



# ==============================================================================
# Copyright 2022 Luca Della Libera.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Negative log unnormalized posterior loss."""

from typing import Any

from torch import Tensor
from torch.nn.modules import loss


__all__ = [
    "NLUPLoss",
]


class NLUPLoss(loss._Loss):
    """Negative log unnormalized posterior loss.

    References
    ----------
    .. [1] T. Chen, E. B. Fox, and C. Guestrin.
           "Stochastic Gradient Hamiltonian Monte Carlo".
           In: ICML. 2014, pp. 1683-1691.
           URL: https://arxiv.org/abs/1402.4102
    .. [2] Q. Liu and D. Wang.
           "Stein Variational Gradient Descent: A General Purpose Bayesian Inference Algorithm".
           In: Advances in Neural Information Processing Systems. 2016, pp. 2378-2386.
           URL: https://arxiv.org/abs/1608.04471

    Examples
    --------
    >>> import torch
    >>>
    >>> from bayestorch.losses import NLUPLoss
    >>>
    >>>
    >>> num_mc_samples = 5
    >>> num_train_batches = 100
    >>> batch_size = 256
    >>> log_likelihoods = torch.rand(num_mc_samples, batch_size)
    >>> log_priors = torch.rand(num_mc_samples)
    >>> criterion = NLUPLoss(reduction="none")
    >>> loss = criterion(log_likelihoods, log_priors, num_train_batches)

    """

    # override
    def __init__(self, reduction: "str" = "mean", **kwargs: "Any") -> "None":
        """Initialize the object.

        Parameters
        ----------
        reduction:
            The reduction to apply to the output. Must be one of the following:
            - "none": no reduction is applied;
            - "sum": the output is summed;
            - "mean": the output sum is divided by the number of elements.

        Raises
        ------
        ValueError
            If an invalid argument value is given.

        """
        if reduction not in ["none", "sum", "mean"]:
            raise ValueError(
                f"`reduction` ({reduction}) must be one of {['none', 'sum', 'mean']}"
            )
        super().__init__(reduction=reduction, **kwargs)

    # override
    def forward(
        self,
        log_likelihoods: "Tensor",
        log_priors: "Tensor",
        log_prior_weight: "float",
    ) -> "Tensor":
        """Forward pass.

        In the following, let `B = {B_1, ..., B_k}` denote
        the batch shape and `N` the number of Monte Carlo
        samples.

        Parameters
        ----------
        log_likelihoods:
            The log likelihoods, shape: ``[N, *B]``.
        log_priors:
            The log priors, shape: ``[N]``.
        log_prior_weight:
            The log prior weight (`1 / M` in the literature).
            According to reference [1], it counterbalances the bias deriving
            from summing the log likelihood over a single batch of data instead
            of over the entire dataset. It is often set equal to the number of
            training batches. More generally, it controls the strength of the
            regularization provided by the log prior term and its optimal
            value depends on factors such as model and dataset size.

        Returns
        -------
            The loss, shape: ``[N]`` if `reduction` initialization
            argument is equal to "none", ``[1]`` otherwise.

        """
        if log_likelihoods.ndim > 1:
            # Sum along batch dimensions if any
            log_likelihoods = log_likelihoods.flatten(start_dim=1).sum(dim=-1)
        result = -(log_likelihoods + log_prior_weight * log_priors)
        if self.reduction == "none":
            return result
        elif self.reduction == "sum":
            return result.sum()
        elif self.reduction == "mean":
            return result.mean()



# ==============================================================================
# Copyright 2022 Luca Della Libera.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Version according to SemVer versioning system (https://semver.org/)."""


__all__ = [
    "VERSION",
]


_MAJOR = "0"  # Major version to increment in case of incompatible API changes

_MINOR = (
    "0"  # Minor version to increment in case of backward compatible new functionality
)

_PATCH = "1"  # Patch version to increment in case of backward compatible bug fixes

VERSION = f"{_MAJOR}.{_MINOR}.{_PATCH}"
"""The package version."""



# ==============================================================================
# Copyright 2022 Luca Della Libera.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Stochastic gradient Langevin dynamics optimizer."""

from typing import Any, Dict, Iterable, Union

import torch
from torch import Tensor
from torch.optim import Optimizer


__all__ = [
    "SGLD",
]


# Adapted from:
# https://github.com/JavierAntoran/Bayesian-Neural-Networks/blob/1f867a5bcbd1abfecede99807eb0b5f97ed8be7c/src/Stochastic_Gradient_Langevin_Dynamics/optimizers.py#L1
class SGLD(Optimizer):
    """Stochastic gradient Langevin dynamics optimizer.

    The optimization parameters are viewed as a posterior
    sample under stochastic gradient Langevin dynamics with
    noise rescaled in each dimension according to RMSProp.

    References
    ----------
    .. [1] C. Li, C. Chen, D. Carlson, and L. Carin.
           "Preconditioned Stochastic Gradient Langevin Dynamics for Deep Neural Networks".
           In: AAAI. 2016, pp. 1788-1794.
           URL: https://arxiv.org/abs/1512.07666

    Examples
    --------
    >>> import torch
    >>> from torch import nn
    >>>
    >>> from bayestorch.optimizers import SGLD
    >>>
    >>>
    >>> batch_size = 10
    >>> in_features = 4
    >>> out_features = 2
    >>> model = nn.Linear(in_features, out_features)
    >>> optimizer = SGLD(model.parameters())
    >>> input = torch.rand(batch_size, in_features)
    >>> output = model(input)
    >>> loss = output.sum()
    >>> loss.backward()
    >>> optimizer.step()

    """

    # override
    def __init__(
        self,
        params: "Union[Iterable[Tensor], Iterable[Dict[str, Any]]]",
        lr: "float" = 1e-2,
        num_burn_in_steps: "int" = 3000,
        precondition_decay_rate: "float" = 0.95,
        epsilon: "float" = 1e-8,
    ) -> "None":
        """Initialize the object.

        Parameters
        ----------
        params:
            The parameters to optimize.
        lr:
            The learning rate.
        num_burn_in_steps:
            The number of steps for which gradient
            statistics are collected to update the
            preconditioner before starting to draw
            noisy samples.
        precondition_decay_rate:
            The exponential decay rate for rescaling the
            preconditioner according to RMSProp. Should
            be close to 1 to approximate sampling from
            the posterior.
        epsilon:
            The term for improving numerical stability.

        Raises
        ------
        ValueError
            If an invalid argument value is given.

        """
        if lr < 0.0:
            raise ValueError(f"`lr` ({lr}) must be in the interval [0, inf)")
        if num_burn_in_steps < 0 or not float(num_burn_in_steps).is_integer():
            raise ValueError(
                f"`num_burn_in_steps` ({num_burn_in_steps}) must be in the integer interval [0, inf)"
            )
        if precondition_decay_rate < 0.0 or precondition_decay_rate > 1.0:
            raise ValueError(
                f"`precondition_decay_rate` ({precondition_decay_rate}) must be in the interval [0, 1]"
            )
        if epsilon <= 0.0:
            raise ValueError(f"`epsilon` ({epsilon}) must be in the interval (0, inf)")

        defaults = {
            "lr": lr,
            "num_burn_in_steps": int(num_burn_in_steps),
            "precondition_decay_rate": precondition_decay_rate,
            "epsilon": epsilon,
        }
        super().__init__(params, defaults)

    # override
    @torch.no_grad()
    def step(self) -> "None":
        for group in self.param_groups:
            for param in group["params"]:
                if param.grad is None:
                    continue

                state = self.state[param]

                # State initialization
                if not state:
                    state["iteration"] = 0
                    state["momentum"] = torch.ones_like(param)

                lr = group["lr"]
                num_burn_in_steps = group["num_burn_in_steps"]
                precondition_decay_rate = group["precondition_decay_rate"]
                epsilon = group["epsilon"]

                iteration = state["iteration"]
                momentum = state["momentum"]

                iteration += 1
                grad = param.grad

                # Momentum update
                momentum += (1.0 - precondition_decay_rate) * (grad**2 - momentum)

                # Burn-in steps
                if iteration <= num_burn_in_steps:
                    stddev = torch.zeros_like(param)
                else:
                    stddev = 1.0 / torch.full_like(param, lr).sqrt()

                # Draw random sample
                preconditioner = 1.0 / (momentum + epsilon).sqrt()
                mean = 0.5 * preconditioner * grad
                stddev *= preconditioner.sqrt()
                sample = torch.normal(mean, stddev)

                # Parameter update
                param += -lr * sample



# ==============================================================================
# Copyright 2022 Luca Della Libera.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Stochastic gradient Hamiltonian Monte Carlo optimizer."""

from typing import Any, Dict, Iterable, Union

import torch
from torch import Tensor
from torch.optim import Optimizer


__all__ = [
    "SGHMC",
]


# Adapted from:
# https://github.com/JavierAntoran/Bayesian-Neural-Networks/blob/1f867a5bcbd1abfecede99807eb0b5f97ed8be7c/src/Stochastic_Gradient_HMC_SA/optimizers.py#L1
class SGHMC(Optimizer):
    """Stochastic gradient Hamiltonian Monte Carlo optimizer.

    A burn-in procedure is used to adapt the hyperparameters
    during the initial stages of sampling.

    References
    ----------
    .. [1] T. Chen, E. B. Fox, and C. Guestrin.
           "Stochastic Gradient Hamiltonian Monte Carlo".
           In: ICML. 2014, pp. 1683-1691.
           URL: https://arxiv.org/abs/1402.4102

    Examples
    --------
    >>> import torch
    >>> from torch import nn
    >>>
    >>> from bayestorch.optimizers import SGHMC
    >>>
    >>>
    >>> batch_size = 10
    >>> in_features = 4
    >>> out_features = 2
    >>> model = nn.Linear(in_features, out_features)
    >>> optimizer = SGHMC(model.parameters())
    >>> input = torch.rand(batch_size, in_features)
    >>> output = model(input)
    >>> loss = output.sum()
    >>> loss.backward()
    >>> optimizer.step()

    """

    # override
    def __init__(
        self,
        params: "Union[Iterable[Tensor], Iterable[Dict[str, Any]]]",
        lr: "float" = 1e-2,
        num_burn_in_steps: "int" = 3000,
        momentum_decay: "float" = 0.05,
        grad_noise: "float" = 0.0,
        epsilon: "float" = 1e-16,
    ) -> "None":
        """Initialize the object.

        Parameters
        ----------
        params:
            The parameters to optimize.
        lr:
            The learning rate.
        num_burn_in_steps:
            The number of burn-in steps. At each step,
            the optimizer hyperparameters are adapted
            to decrease the error.
        momentum_decay:
            The momentum decay per timestep.
        grad_noise:
            The constant per-parameter gradient
            noise used for sampling.
        epsilon:
            The term for improving numerical stability.

        Raises
        ------
        ValueError
            If an invalid argument value is given.

        """
        if lr < 0.0:
            raise ValueError(f"`lr` ({lr}) must be in the interval [0, inf)")
        if num_burn_in_steps < 0 or not float(num_burn_in_steps).is_integer():
            raise ValueError(
                f"`num_burn_in_steps` ({num_burn_in_steps}) must be in the integer interval [0, inf)"
            )
        if momentum_decay < 0.0:
            raise ValueError(
                f"`momentum_decay` ({momentum_decay}) must be in the interval [0, inf)"
            )
        if grad_noise < 0.0:
            raise ValueError(
                f"`grad_noise` ({grad_noise}) must be in the interval [0, inf)"
            )
        if epsilon <= 0.0:
            raise ValueError(f"`epsilon` ({epsilon}) must be in the interval (0, inf)")

        defaults = {
            "lr": lr,
            "num_burn_in_steps": int(num_burn_in_steps),
            "momentum_decay": momentum_decay,
            "grad_noise": grad_noise,
            "epsilon": epsilon,
        }
        super().__init__(params, defaults)

    # override
    @torch.no_grad()
    def step(self) -> "None":
        for group in self.param_groups:
            for param in group["params"]:
                if param.grad is None:
                    continue

                state = self.state[param]

                # State initialization
                if not state:
                    state["iteration"] = 0
                    state["tau"] = torch.ones_like(param)
                    state["g"] = torch.ones_like(param)
                    state["v_hat"] = torch.ones_like(param)
                    state["momentum"] = torch.zeros_like(param)

                lr = group["lr"]
                num_burn_in_steps = group["num_burn_in_steps"]
                momentum_decay = group["momentum_decay"]
                grad_noise = group["grad_noise"]
                epsilon = group["epsilon"]

                iteration = state["iteration"]
                tau = state["tau"]
                g = state["g"]
                v_hat = state["v_hat"]
                momentum = state["momentum"]

                iteration += 1
                grad = param.grad
                r = 1.0 / (tau + 1.0)
                m_inv = 1.0 / v_hat.sqrt()

                # Burn-in steps
                if iteration <= num_burn_in_steps:
                    tau += 1.0 - tau * (g**2 / v_hat)
                    g += (grad - g) * r
                    v_hat += (grad**2 - v_hat) * r

                # Draw random sample
                grad_noise_var = (
                    2.0 * (lr**2) * momentum_decay * m_inv
                    - 2.0 * (lr**3) * (m_inv**2) * grad_noise
                    - (lr**4)
                )
                stddev = grad_noise_var.clamp(min=epsilon).sqrt()
                sample = torch.normal(0.0, stddev)

                # Parameter update
                momentum += sample - lr**2 * m_inv * grad - momentum_decay * momentum
                param += momentum



# ==============================================================================
# Copyright 2022 Luca Della Libera.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Radial basis function Stein kernel."""

import math
from typing import Tuple

import torch
from torch import Tensor


__all__ = [
    "RBFSteinKernel",
]


class RBFSteinKernel:
    """Radial basis function kernel to use in Stein
    variational gradient descent.

    The bandwidth of the kernel is chosen from the particles
    using a simple heuristic as in reference [1].

    References
    ----------
    .. [1] Q. Liu and D. Wang.
           "Stein Variational Gradient Descent: A General Purpose Bayesian Inference Algorithm".
           In: Advances in Neural Information Processing Systems. 2016, pp. 2378-2386.
           URL: https://arxiv.org/abs/1608.04471

    Examples
    --------
    >>> import torch
    >>>
    >>> from bayestorch.kernels import RBFSteinKernel
    >>>
    >>>
    >>> num_particles = 5
    >>> particle_size = 1000
    >>> kernel = RBFSteinKernel()
    >>> particles = torch.rand(num_particles, particle_size)
    >>> kernels, kernel_grads = kernel(particles)

    """

    @torch.no_grad()
    def __call__(self, particles: "Tensor") -> "Tuple[Tensor, Tensor]":
        """Compute the kernels and the kernel gradients.

        In the following, let `N` denote the number
        of particles and `D` the particle size.

        Parameters
        ----------
        particles:
            The particles, shape: ``[N, D]``.

        Returns
        -------
            - The kernels, shape: ``[N, N]``;
            - the kernel gradients with respect
              to the particles, shape: ``[N, D]``.

        """
        num_particles = particles.shape[0]
        return self._forward(particles, num_particles)

    @staticmethod
    @torch.jit.script
    def _forward(particles: "Tensor", num_particles: "int") -> "Tuple[Tensor, Tensor]":
        deltas = torch.cdist(particles, particles)
        squared_deltas = deltas**2
        bandwidth = squared_deltas.median() / math.log(num_particles)
        log_kernels = -squared_deltas / bandwidth
        kernels = log_kernels.exp()
        kernel_grads = (
            (2 / bandwidth) * (kernels.sum(dim=-1).diag() - kernels) @ particles
        )
        return kernels, kernel_grads

    def __repr__(self) -> "str":
        return f"{type(self).__name__}()"



# ==============================================================================
# Copyright 2022 Luca Della Libera.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Model utilities."""

from typing import Dict, Sequence, TypeVar, Union

import torch
from torch import Tensor


__all__ = [
    "nested_stack",
]


_T = TypeVar("_T")

_Nested = Union[_T, Sequence[_T], Dict[str, _T]]


def nested_stack(
    inputs: "Sequence[_Nested[Tensor]]", dim: "int" = 0
) -> "_Nested[Tensor]":
    """Stack a sequence of possibly nested tensors
    along a dimension.

    Parameters
    ----------
    inputs:
        The sequence of possibly nested tensors.
    dim:
        The dimension.

    Examples
    --------
    >>> import torch
    >>>
    >>> from bayestorch.models.utils import nested_stack
    >>>
    >>>
    >>> num_outputs = 4
    >>> inputs = [
    ...    {"a": [torch.rand(2, 3), torch.rand(3, 5)], "b": torch.rand(1, 2)}
    ...    for _ in range(num_outputs)
    ... ]
    >>> outputs = nested_stack(inputs)

    """
    first_input = inputs[0]
    if isinstance(first_input, Tensor):
        return torch.stack(inputs, dim=dim)
    if first_input is None:
        return None
    if isinstance(first_input, dict):
        return type(first_input)(
            (k, nested_stack([output[k] for output in inputs])) for k in first_input
        )
    return type(first_input)(map(nested_stack, zip(*inputs)))



# ==============================================================================
# Copyright 2022 Luca Della Libera.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Variational posterior model."""

import copy
import logging
from typing import Any, Callable, Dict, Iterator, List, Optional, Tuple, TypeVar

from torch import Tensor, nn
from torch.distributions import Distribution, kl_divergence
from torch.nn import Module, Parameter

from bayestorch.models.prior_model import PriorModel
from bayestorch.models.utils import nested_stack


__all__ = [
    "VariationalPosteriorModel",
]


_T = TypeVar("_T", bound="VariationalPosteriorModel")

_LOGGER = logging.getLogger(__name__)


class VariationalPosteriorModel(PriorModel):
    """Bayesian model that defines a prior and a variational
    posterior over its parameters.

    References
    ----------
    .. [1] C. Blundell, J. Cornebise, K. Kavukcuoglu, and D. Wierstra.
           "Weight Uncertainty in Neural Networks".
           In: ICML. 2015, pp. 1613-1622.
           URL: https://arxiv.org/abs/1505.05424

    Examples
    --------
    >>> import torch
    >>> from torch import nn
    >>>
    >>> from bayestorch.distributions import LogScaleNormal, SoftplusInvScaleNormal
    >>> from bayestorch.models import VariationalPosteriorModel
    >>>
    >>>
    >>> num_mc_samples = 5
    >>> batch_size = 10
    >>> in_features = 4
    >>> out_features = 2
    >>> model = nn.Linear(in_features, out_features)
    >>> num_parameters = sum(parameter.numel() for parameter in model.parameters())
    >>> model = VariationalPosteriorModel(
    ...     model,
    ...     prior_builder=LogScaleNormal,
    ...     prior_kwargs={
    ...         "loc": torch.zeros(num_parameters),
    ...         "log_scale": torch.full((num_parameters,), -1.0),
    ...     },
    ...     posterior_builder=SoftplusInvScaleNormal,
    ...     posterior_kwargs={
    ...         "loc": torch.zeros(num_parameters, requires_grad=True),
    ...         "softplus_inv_scale": torch.full((num_parameters,), -7.0, requires_grad=True),
    ...     },
    ... )
    >>> input = torch.rand(batch_size, in_features)
    >>> outputs, kl_divs = model(
    ...     input,
    ...     num_mc_samples=num_mc_samples,
    ...     exact_kl_div=False,
    ... )

    """

    posterior: "Distribution"
    """The posterior distribution."""

    # override
    def __init__(
        self,
        model: "Module",
        prior_builder: "Callable[..., Distribution]",
        prior_kwargs: "Dict[str, Any]",
        posterior_builder: "Callable[..., Distribution]",
        posterior_kwargs: "Dict[str, Any]",
    ) -> "None":
        """Initialize the object.

        Parameters
        ----------
        model:
            The model.
        prior_builder:
            The prior builder, i.e. a callable that receives
            keyword arguments and returns a prior.
        prior_kwargs:
            The keyword arguments to pass to the prior builder.
            Tensor arguments are internally registered as
            parameters if their `requires_grad` attribute
            is True, as persistent buffers otherwise.
        posterior_builder:
            The posterior builder, i.e. a callable that receives
            keyword arguments and returns a posterior.
        posterior_kwargs:
            The keyword arguments to pass to the posterior builder.
            Tensor arguments are internally registered as
            parameters if their `requires_grad` attribute
            is True, as persistent buffers otherwise.

        Raises
        ------
        ValueError
            If an invalid argument value is given.

        """
        super().__init__(model, prior_builder, prior_kwargs)
        self.posterior_builder = posterior_builder
        self.posterior_kwargs = posterior_kwargs = {
            k: v for k, v in posterior_kwargs.items()
        }  # Avoid side effects

        # Build prior
        self.posterior = self._build_distribution(
            "posterior", posterior_builder, posterior_kwargs
        )

        # Log Kullback-Leibler divergence warning only once
        self._log_kl_div_warning = True

    # override
    def named_parameters(
        self, prefix: "str" = "", recurse: "bool" = True
    ) -> "Iterator[Tuple[str, Parameter]]":
        return super().named_parameters(recurse=False)

    # override
    def state_dict(
        self, *args: "Any", **kwargs: "Any"
    ) -> "Dict[str, Optional[Tensor]]":
        tmp = super().state_dict(keep_vars=True)
        # Remove underlying model parameters but keep buffers
        for key in list(tmp.keys()):
            if ("prior" not in key and "posterior" not in key) and tmp[
                key
            ].requires_grad:
                tmp.pop(key)
        result = super().state_dict(*args, **kwargs)
        result = {k: v for k, v in result.items() if k in tmp}
        return result

    # override
    def load_state_dict(
        self, *args: "Any", strict: "bool" = True, **kwargs: "Any"
    ) -> "Tuple[List[str], List[str]]":
        return super().load_state_dict(*args, strict=False, **kwargs)

    # override
    def to(self, *args: "Any", **kwargs: "Any") -> "_T":
        super().to(*args, **kwargs)

        # Rebuild posterior with parameters on new device
        self.prior = self._build_distribution(
            "prior", self.prior_builder, self.prior_kwargs
        )

        return self

    # override
    def forward(
        self,
        *args: "Any",
        num_mc_samples: "int" = 1,
        exact_kl_div: "bool" = False,
        **kwargs: "Any",
    ) -> "Tuple[Any, Tensor]":
        """Forward pass.

        In the following, let `N` denote the number of Monte Carlo samples,
        `B = {B_1, ..., B_k}` the batch shape, and `O = {O_1, ..., O_m}`
        the shape of a leaf value of the underlying model output (can be
        a nested tensor).

        Parameters
        ----------
        args:
            The positional arguments to pass to the underlying model.
        num_mc_samples:
            The number of Monte Carlo samples.
        exact_kl_div:
            True to use the exact Kullback-Leibler divergence between
            posterior and prior (if a closed-form expression exists),
            False to use Monte Carlo approximation.
            For consistency, the scalar returned by the closed-form
            Kullback-Leibler divergence is expanded to a tensor of
            shape: ``[N]``.
        kwargs:
            The keyword arguments to pass to the underlying model.

        Returns
        -------
            - The outputs, shape of a leaf value: ``[N, *B, *O]``;
            - the Kullback-Leibler divergences between posterior
              and prior, shape: ``[N]``.

        Raises
        ------
        ValueError
            If an invalid argument value is given.

        """
        if num_mc_samples < 1 or not float(num_mc_samples).is_integer():
            raise ValueError(
                f"`num_mc_samples` ({num_mc_samples}) must be in the integer interval [1, inf)"
            )

        # Sample new particles
        new_particles = self.posterior.rsample((num_mc_samples,))

        # Compute Kullback-Leibler divergences
        kl_divs = None
        if exact_kl_div:
            try:
                kl_divs = kl_divergence(self.posterior, self.prior).expand(
                    num_mc_samples
                )
            except NotImplementedError:
                kl_divs = None
                if self._log_kl_div_warning:
                    _LOGGER.warning(
                        "Could not compute exact Kullback-Leibler divergence, "
                        "reverting to Monte Carlo approximation"
                    )
                    self._log_kl_div_warning = False
        if kl_divs is None:
            log_posteriors = self.posterior.log_prob(new_particles)
            log_priors = self.prior.log_prob(new_particles)
            kl_divs = log_posteriors - log_priors

        # Replicate model (one replica for each Monte Carlo sample)
        models = nn.ModuleList(
            [self.model]
            + [copy.deepcopy(self.model) for _ in range(num_mc_samples - 1)]
        )

        # Inject sampled particles
        new_particles = new_particles.flatten()
        start_idx = 0
        for parameter in models.parameters():
            num_elements = parameter.numel()
            new_parameter = new_particles[
                start_idx : start_idx + num_elements
            ].reshape_as(parameter)
            parameter.detach_().requires_grad_(False).copy_(new_parameter)
            start_idx += num_elements

        # Forward pass
        outputs = [model(*args, **kwargs) for model in models]

        return nested_stack(outputs), kl_divs

    # override
    def __repr__(self) -> "str":
        return (
            f"{type(self).__name__}"
            f"(model: {self.model}, "
            f"prior: {self.prior}, "
            f"posterior: {self.posterior})"
        )



# ==============================================================================
# Copyright 2022 Luca Della Libera.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Setup script."""


################## Install setup requirements ##################
def _preinstall_requirement(requirement, options=None):
    import subprocess

    args = ["pip", "install", requirement, *(options or [])]
    return_code = subprocess.call(args)
    if return_code != 0:
        raise RuntimeError(f"{requirement} installation failed")


for requirement in ["setuptools>=58.0.4", "wheel>=0.37.1"]:
    _preinstall_requirement(requirement)
################################################################


import os  # noqa: E402

from setuptools import find_packages, setup  # noqa: E402


_ROOT_DIR = os.path.dirname(os.path.realpath(__file__))

with open(os.path.join(_ROOT_DIR, "bayestorch", "version.py")) as f:
    tmp = {}
    exec(f.read(), tmp)
    _VERSION = tmp["VERSION"]

with open(os.path.join(_ROOT_DIR, "README.md"), encoding="utf-8") as f:
    _README = f.read()

setup(
    name="bayestorch",
    version=_VERSION,
    description="Bayesian deep learning library for fast prototyping based on PyTorch",
    long_description=_README,
    long_description_content_type="text/markdown",
    author="Luca Della Libera",
    author_email="luca.dellalib@gmail.com",
    url="https://github.com/lucadellalib/bayestorch",
    packages=find_packages(),
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Console",
        "Environment :: GPU :: NVIDIA CUDA",
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Typing :: Typed",
    ],
    license="Apache License 2.0",
    keywords=["Bayesian deep learning", "PyTorch"],
    platforms=["OS Independent"],
    include_package_data=True,
    install_requires=["torch>=1.6.0"],
    extras_require={
        "test": [
            "pytest>=5.4.3",
            "pytest-cov>=2.9.0",
        ],
        "dev": [
            "black>=22.3.0",
            "cibuildwheel>=2.3.1",
            "flake8>=3.8.3",
            "flake8-bugbear>=20.1.4",
            "isort>=5.4.2",
            "pre-commit>=2.6.0",
            "pre-commit-hooks>=3.2.0",
            "pytest>=5.4.3",
            "pytest-cov>=2.9.0",
            "twine>=3.3.0",
        ],
    },
    python_requires=">=3.6",
)



# ==============================================================================
# Copyright 2022 Luca Della Libera.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Particle posterior model."""

import copy
from typing import Any, Callable, Dict, Tuple

from torch import Tensor, nn
from torch.distributions import Distribution
from torch.nn import Module, ModuleList

from bayestorch.models.prior_model import PriorModel
from bayestorch.models.utils import nested_stack


__all__ = [
    "ParticlePosteriorModel",
]


class ParticlePosteriorModel(PriorModel):
    """Bayesian model that defines a prior and a particle-based
    posterior over its parameters.

    References
    ----------
    .. [1] Q. Liu and D. Wang.
           "Stein Variational Gradient Descent: A General Purpose Bayesian Inference Algorithm".
           In: Advances in Neural Information Processing Systems. 2016, pp. 2378-2386.
           URL: https://arxiv.org/abs/1608.04471

    Examples
    --------
    >>> import torch
    >>> from torch import nn
    >>>
    >>> from bayestorch.distributions import LogScaleNormal
    >>> from bayestorch.models import ParticlePosteriorModel
    >>>
    >>>
    >>> num_particles = 5
    >>> batch_size = 10
    >>> in_features = 4
    >>> out_features = 2
    >>> model = nn.Linear(in_features, out_features)
    >>> num_parameters = sum(parameter.numel() for parameter in model.parameters())
    >>> model = ParticlePosteriorModel(
    ...     model,
    ...     prior_builder=LogScaleNormal,
    ...     prior_kwargs={
    ...         "loc": torch.zeros(num_parameters),
    ...         "log_scale": torch.full((num_parameters,), -1.0),
    ...     },
    ...     num_particles=num_particles,
    ... )
    >>> input = torch.rand(batch_size, in_features)
    >>> outputs, log_priors = model(input)

    """

    models: "ModuleList"
    """The model replicas (one for each particle)."""

    # override
    def __init__(
        self,
        model: "Module",
        prior_builder: "Callable[..., Distribution]",
        prior_kwargs: "Dict[str, Any]",
        num_particles: "int" = 10,
    ) -> "None":
        """Initialize the object.

        Parameters
        ----------
        model:
            The model.
        prior_builder:
            The prior builder, i.e. a callable that receives
            keyword arguments and returns a prior.
        prior_kwargs:
            The keyword arguments to pass to the prior builder.
            Tensor arguments are internally registered as
            parameters if their `requires_grad` attribute
            is True, as persistent buffers otherwise.
        num_particles:
            The number of particles.

        Raises
        ------
        ValueError
            If an invalid argument value is given.

        """
        if num_particles < 1 or not float(num_particles).is_integer():
            raise ValueError(
                f"`num_particles` ({num_particles}) must be in the integer interval [1, inf)"
            )

        super().__init__(model, prior_builder, prior_kwargs)
        self.num_particles = num_particles

        # Replicate model (one replica for each particle)
        self.models = ModuleList(
            [model] + [copy.deepcopy(model) for _ in range(num_particles - 1)]
        )

        for model in self.models:
            # Sample new particle
            new_particle = self.prior.sample()

            # Inject sampled particle
            start_idx = 0
            for parameter in model.parameters():
                num_elements = parameter.numel()
                new_parameter = new_particle[
                    start_idx : start_idx + num_elements
                ].reshape_as(parameter)
                parameter.detach_().requires_grad_(False).copy_(
                    new_parameter
                ).requires_grad_(True)
                start_idx += num_elements

    # override
    def forward(self, *args: "Any", **kwargs: "Any") -> "Tuple[Any, Tensor]":
        """Forward pass.

        In the following, let `N` denote the number of particles,
        `B = {B_1, ..., B_k}` the batch shape, and `O = {O_1, ..., O_m}`
        the shape of a leaf value of the underlying model output (can be
        a nested tensor).

        Parameters
        ----------
        args:
            The positional arguments to pass to the underlying model.
        kwargs:
            The keyword arguments to pass to the underlying model.

        Returns
        -------
            - The outputs, shape of a leaf value: ``[N, *B, *O]``;
            - the log priors, shape: ``[N]``.

        """
        # Extract particles
        particles = nn.utils.parameters_to_vector(self.models.parameters()).reshape(
            self.num_particles, -1
        )

        # Compute log priors
        log_priors = self.prior.log_prob(particles)

        # Forward pass
        outputs = [model(*args, **kwargs) for model in self.models]

        return nested_stack(outputs), log_priors

    # override
    def __repr__(self) -> "str":
        return (
            f"{type(self).__name__}"
            f"(model: {self.model}, "
            f"prior: {self.prior}, "
            f"num_particles: {self.num_particles})"
        )



# ==============================================================================
# Copyright 2022 Luca Della Libera.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Log scale normal distribution."""

from typing import Optional, Union

from torch import Size, Tensor
from torch.distributions import Normal, constraints


__all__ = [
    "LogScaleNormal",
]


class LogScaleNormal(Normal):
    """Normal distribution parameterized by location
    and log scale parameters.

    Scale parameter is computed as `exp(log_scale)`.

    Examples
    --------
    >>> from bayestorch.distributions import LogScaleNormal
    >>>
    >>>
    >>> loc = 0.0
    >>> log_scale = -1.0
    >>> distribution = LogScaleNormal(loc, log_scale)

    """

    arg_constraints = {
        "loc": constraints.real,
        "log_scale": constraints.real,
    }  # override

    # override
    def __init__(
        self,
        loc: "Union[int, float, Tensor]",
        log_scale: "Union[int, float, Tensor]",
        validate_args: "Optional[bool]" = None,
    ) -> "None":
        super().__init__(loc, log_scale, validate_args)

    # override
    @property
    def scale(self) -> "Tensor":
        return self.log_scale.exp()

    # override
    @scale.setter
    def scale(self, value: "Tensor") -> "None":
        self.log_scale = value

    # override
    def expand(
        self,
        batch_shape: "Size" = Size(),  # noqa: B008
        _instance: "Optional[LogScaleNormal]" = None,
    ) -> "LogScaleNormal":
        new = self._get_checked_instance(LogScaleNormal, _instance)
        loc = self.loc.expand(batch_shape)
        log_scale = self.log_scale.expand(batch_shape)
        super(LogScaleNormal, new).__init__(loc, log_scale, validate_args=False)
        new._validate_args = self._validate_args
        return new



# ==============================================================================
# Copyright 2022 Luca Della Libera.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Stein variational gradient descent preconditioner."""

from typing import Any, Callable, Dict, Iterable, Tuple, Union

import torch
from torch import Tensor, nn
from torch.optim import Optimizer


__all__ = [
    "SVGD",
]


class SVGD(Optimizer):
    """Stein variational gradient descent preconditioner.

    References
    ----------
    .. [1] Q. Liu and D. Wang.
           "Stein Variational Gradient Descent: A General Purpose Bayesian Inference Algorithm".
           In: Advances in Neural Information Processing Systems. 2016, pp. 2378-2386.
           URL: https://arxiv.org/abs/1608.04471

    Examples
    --------
    >>> import torch
    >>> from torch import nn
    >>>
    >>> from bayestorch.kernels import RBFSteinKernel
    >>> from bayestorch.preconditioners import SVGD
    >>>
    >>>
    >>> num_particles = 5
    >>> batch_size = 10
    >>> in_features = 4
    >>> out_features = 2
    >>> models = nn.ModuleList(
    ...     [nn.Linear(in_features, out_features) for _ in range(num_particles)]
    ... )
    >>> kernel = RBFSteinKernel()
    >>> preconditioner = SVGD(models.parameters(), kernel, num_particles)
    >>> input = torch.rand(batch_size, in_features)
    >>> outputs = torch.cat([model(input) for model in models])
    >>> loss = outputs.sum()
    >>> loss.backward()
    >>> preconditioner.step()

    """

    # override
    def __init__(
        self,
        params: "Union[Iterable[Tensor], Iterable[Dict[str, Any]]]",
        kernel: "Callable[[Tensor], Tuple[Tensor, Tensor]]",
        num_particles: "int",
    ) -> "None":
        """Initialize the object.

        Parameters
        ----------
        params:
            The parameters to precondition. The total number of
            parameters must be a multiple of `num_particles`.
        kernel:
            The kernel, i.e. a callable that receives as an argument the
            particles and returns the corresponding kernels and kernel
            gradients.
        num_particles:
            The number of particles.

        Raises
        ------
        ValueError
            If an invalid argument value is given.

        """
        super().__init__(params, {"kernel": kernel, "num_particles": num_particles})
        self.kernel = kernel
        self.num_particles = num_particles = int(num_particles)

        # Check consistency between number of parameters
        # and number of particles for each group
        for group in self.param_groups:
            params = group["params"]
            num_particles = group["num_particles"]

            if num_particles < 1 or not float(num_particles).is_integer():
                raise ValueError(
                    f"`num_particles` ({num_particles}) must be in the integer interval [1, inf)"
                )

            # Extract particles
            particles = nn.utils.parameters_to_vector(params)

            if particles.numel() % num_particles != 0:
                raise ValueError(
                    f"Total number of parameters ({particles.numel()}) must "
                    f"be a multiple of `num_particles` ({num_particles})"
                )

    # override
    @torch.no_grad()
    def step(self) -> "None":
        for group in self.param_groups:
            params = group["params"]
            kernel = group["kernel"]
            num_particles = group["num_particles"]

            # Extract particles
            particles = nn.utils.parameters_to_vector(params).reshape(num_particles, -1)

            # Extract particle gradients
            particle_grads = []
            for param in params:
                grad = param.grad
                if grad is None:
                    raise RuntimeError("Gradient of some parameters is None")
                particle_grads.append(grad)
            particle_grads = nn.utils.parameters_to_vector(particle_grads).reshape(
                num_particles, -1
            )

            # Compute kernels and kernel gradients
            kernels, kernel_grads = kernel(particles)

            # Attractive gradients (already divided by `num_particles` => use `NLUPLoss` with reduction="mean")
            particle_grads = kernels @ particle_grads

            # Repulsive gradients
            particle_grads -= kernel_grads / num_particles

            # Flatten
            particle_grads = particle_grads.flatten()

            # Inject particle gradients
            start_idx = 0
            for param in params:
                num_elements = param.numel()
                param.grad = particle_grads[
                    start_idx : start_idx + num_elements
                ].reshape_as(param)
                start_idx += num_elements



# ==============================================================================
# Copyright 2022 Luca Della Libera.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Evidence lower bound (a.k.a. variational free energy) loss."""

from typing import Any

from torch import Tensor
from torch.nn.modules import loss


__all__ = [
    "ELBOLoss",
]


class ELBOLoss(loss._Loss):
    """Evidence lower bound (a.k.a. variational free energy) loss.

    References
    ----------
    .. [1] C. Blundell, J. Cornebise, K. Kavukcuoglu, and D. Wierstra.
           "Weight Uncertainty in Neural Networks".
           In: ICML. 2015, pp. 1613-1622.
           URL: https://arxiv.org/abs/1505.05424

    Examples
    --------
    >>> import torch
    >>>
    >>> from bayestorch.losses import ELBOLoss
    >>>
    >>>
    >>> num_mc_samples = 5
    >>> num_train_batches = 100
    >>> batch_size = 256
    >>> log_likelihoods = torch.rand(num_mc_samples, batch_size)
    >>> kl_divs = torch.rand(num_mc_samples)
    >>> criterion = ELBOLoss(reduction="none")
    >>> loss = criterion(log_likelihoods, kl_divs, num_train_batches)

    """

    # override
    def __init__(self, reduction: "str" = "mean", **kwargs: "Any") -> "None":
        """Initialize the object.

        Parameters
        ----------
        reduction:
            The reduction to apply to the output. Must be one of the following:
            - "none": no reduction is applied;
            - "sum": the output is summed;
            - "mean": the output sum is divided by the number of elements.

        Raises
        ------
        ValueError
            If an invalid argument value is given.

        """
        if reduction not in ["none", "sum", "mean"]:
            raise ValueError(
                f"`reduction` ({reduction}) must be one of {['none', 'sum', 'mean']}"
            )
        super().__init__(reduction=reduction, **kwargs)

    # override
    def forward(
        self,
        log_likelihoods: "Tensor",
        kl_divs: "Tensor",
        kl_div_weight: "float",
    ) -> "Tensor":
        """Forward pass.

        In the following, let `B = {B_1, ..., B_k}` denote
        the batch shape and `N` the number of Monte Carlo
        samples.

        Parameters
        ----------
        log_likelihoods:
            The log likelihoods, shape: ``[N, *B]``.
        kl_divs:
            The Kullback-Leibler divergences, shape: ``[N]``.
        kl_div_weight:
            The Kullback-Leibler divergence weight (`1 / M` in the literature).
            According to reference [1], it counterbalances the bias deriving
            from summing the log likelihood over a single batch of data instead
            of over the entire dataset. It is often set equal to the number of
            training batches. More generally, it controls the strength of the
            regularization provided by the Kullback-Leibler divergence term and
            its optimal value depends on factors such as model and dataset size.

        Returns
        -------
            The loss, shape: ``[N]`` if `reduction` initialization
            argument is equal to "none", ``[1]`` otherwise.

        """
        if log_likelihoods.ndim > 1:
            # Sum along batch dimensions if any
            log_likelihoods = log_likelihoods.flatten(start_dim=1).sum(dim=-1)
        result = kl_div_weight * kl_divs - log_likelihoods
        if self.reduction == "none":
            return result
        elif self.reduction == "sum":
            return result.sum()
        elif self.reduction == "mean":
            return result.mean()
